# -*- coding: utf-8 -*-
""" Classical Memory """
import ctypes

class CMem(ctypes.Structure):
    """ Classical Memory

    Attributes
    ----------
    cmem_num : int
        number of the classical register (classical memory size).
    bit_array : list (int)
        bit array of classical memory.

    """
    _fields_ = [
        ('cmem_num', ctypes.c_int),
        ('bit_array', ctypes.c_void_p),
    ]

    def __new__(cls, cmem_num, **kwargs):
        """
        Parameters
        ----------
        cmem_num : int
            number of the classical register.

        Returns
        -------
        cmem : instance (CMem)

        """
        obj = cmem_init(cmem_num)
        cmem = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        return cmem

    def __str__(self):

        return str(self.get_bits())

    @property
    def bits(self):
        """ bit list of classical memory. """
        return self.get_bits()

    def clone(self):
        """
        get the copy of the classical memory.

        Parameters
        ----------
        None

        Returns
        -------
        cmem : instance of CMem
            copy of the original classical memory.

        """
        obj = cmem_copy(self)
        cmem = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return cmem

    def get_bits(self):
        """
        get bit list of the classical memory.

        Parameters
        ----------
        None

        Returns
        -------
        bits : numpy.ndarray (int)
            bits array of the classical memory

        """
        bits = cmem_get_bits(self)
        return bits

    def __del__(self):

        cmem_free(self)

# c-library for qstate
from qlazy.lib.cmem_c import cmem_init, cmem_copy, cmem_get_bits, cmem_free



# -*- coding: utf-8 -*-
""" functions related to probability """
import math

import qlazy.config as cfg

def freq2prob(freq):
    """
    frequency to probability

    Arguments
    ----------
    freq : instance of Counter
        frequency data
        ex) Counter({'00': 53, '11': 47})

    Returns
    -------
    prob : dict
        probability data
        ex) {'00': 0.53, '11': 0.47}

    """
    total = sum(freq.values())
    prob = {k: v/total for k, v in freq.items()}
    return prob

def entropy(prob):
    """
    entropy

    Arguments
    ----------
    prob : dict
        probability data
        ex) {'00': 0.53, '11': 0.47}

    Returns
    -------
    value : float
        entropy, sum of -prob[i] log(prob[i])

    """
    if abs(sum(prob.values()) - 1.0) > cfg.EPS:
        raise ValueError("sum of probabilities is not 1.0")

    value = 0.0
    for v in prob.values():
        if v == 0.0:
            continue
        value -= v * math.log2(v)

    return value

def kl_divergence(prob_0, prob_1):
    """
    KL divergence

    Arguments
    ----------
    prob_0 : dict
        probability data
        ex) {'00': 0.53, '11': 0.47}

    prob_1 : dict
        probability data
        ex) {'00': 0.51, '11': 0.49}

    Returns
    -------
    value : float
        KL divergence, sum of prob_0[i] log(prob_0[i] / prob_1[i])

    """
    if ((abs(sum(prob_0.values()) - 1.0) > cfg.EPS or
         abs(sum(prob_1.values()) - 1.0) > cfg.EPS)):
        raise ValueError("sum of probabilities is not 1.0")

    value = 0.0
    for k, v in prob_0.items():
        if v < 0.0:
            raise ValueError("probability must be positive.")
        if v > 1.0:
            raise ValueError("probability must be less than 1.")
        if v == 0.0:
            continue
        if v > 0.0:
            if k not in prob_1.keys() or prob_1[k] == 0.0:
                value = float('inf')
                break
        value += (v * math.log2(v / prob_1[k]))

    return value

def cross_entropy(prob_0, prob_1):
    """
    cross entropy

    Arguments
    ----------
    prob_0 : dict
        probability data
        ex) {'00': 0.53, '11': 0.47}

    prob_1 : dict
        probability data
        ex) {'00': 0.51, '11': 0.49}

    Returns
    -------
    value : float
        cross entropy, sum of -prob_0[i] log(prob_1[i])

    """
    if ((abs(sum(prob_0.values()) - 1.0) > cfg.EPS or
         abs(sum(prob_1.values()) - 1.0) > cfg.EPS)):
        raise ValueError("sum of probabilities is not 1.0")

    value = 0.0
    for k, v in prob_0.items():
        if v == 0.0:
            continue
        if k not in prob_1.keys():
            continue
        if prob_1[k] == 0.0:
            continue
        value -= v * math.log2(prob_1[k])

    return value



# -*- coding: utf-8 -*-
""" wrapper functions for QState """
import ctypes
from ctypes.util import find_library
from collections import Counter
import pathlib

import qlazy.config as cfg
from qlazy.util import get_lib_ext
from qlazy.Stabilizer import Stabilizer
from qlazy.QCirc import QCirc
from qlazy.CMem import CMem

lib = ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"), mode=ctypes.RTLD_GLOBAL)

def stabilizer_init(gene_num=None, qubit_num=None, seed=None):
    """ initialize Stabilizer object """

    libc.srand(ctypes.c_int(seed))

    stab = None
    c_stab = ctypes.c_void_p(stab)

    lib.stabilizer_init.restype = ctypes.c_int
    lib.stabilizer_init.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                    ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.stabilizer_init(ctypes.c_int(gene_num), ctypes.c_int(qubit_num),
                              ctypes.c_int(seed), c_stab)

    if ret == cfg.FALSE:
        raise ValueError("can't initialize Stabilizer object.")

    return c_stab

def stabilizer_copy(sb):
    """ copy the stabilizer state """

    try:
        stab = None
        c_stab = ctypes.c_void_p(stab)

        lib.stabilizer_copy.restype = ctypes.c_int
        lib.stabilizer_copy.argtypes = [ctypes.POINTER(Stabilizer),
                                        ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.stabilizer_copy(ctypes.byref(sb), c_stab)

        if ret == cfg.FALSE:
            raise ValueError("can't copy the Stabilizer object.")

        return c_stab

    except Exception:
        raise ValueError("can't copy the Stabilizer object.")

def stabilizer_set_pauli_fac(sb, gene_id, pauli_fac):
    """ set pauli factor to the stabilizer state """

    lib.stabilizer_set_pauli_fac.restype = ctypes.c_int
    lib.stabilizer_set_pauli_fac.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                             ctypes.c_int]
    ret = lib.stabilizer_set_pauli_fac(ctypes.byref(sb), ctypes.c_int(gene_id),
                                       ctypes.c_int(pauli_fac))

    if ret == cfg.FALSE:
        raise ValueError("can't set pauli factor to the Stabilizer object.")

    return ret

def stabilizer_get_pauli_fac(sb, gene_id):
    """ get pauli factor from the stabilizer state """

    pauli_fac = cfg.REAL_PLUS
    c_pauli_fac = ctypes.c_int(pauli_fac)

    lib.stabilizer_get_pauli_fac.restype = ctypes.c_int
    lib.stabilizer_get_pauli_fac.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                             ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_get_pauli_fac(ctypes.byref(sb), ctypes.c_int(gene_id),
                                       ctypes.byref(c_pauli_fac))

    if ret == cfg.FALSE:
        raise ValueError("can't get pauli factor from the Stabilizer object.")

    pauli_fac = c_pauli_fac.value

    return pauli_fac

def stabilizer_set_pauli_op(sb, gene_id=None, qubit_id=None, pauli_op=None):
    """ set pauli operator to the stabilizer state """

    lib.stabilizer_set_pauli_op.restype = ctypes.c_int
    lib.stabilizer_set_pauli_op.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                            ctypes.c_int, ctypes.c_int]
    ret = lib.stabilizer_set_pauli_op(ctypes.byref(sb), ctypes.c_int(gene_id),
                                      ctypes.c_int(qubit_id), ctypes.c_int(pauli_op))

    if ret == cfg.FALSE:
        raise ValueError("can't set pauli operator to the Stabilizer object.")

    return ret

def stabilizer_get_pauli_op(sb, gene_id=None, qubit_id=None):
    """ get pauli operator from the stabilizer state """

    pauli_op = cfg.IDENTITY
    c_pauli_op = ctypes.c_int(pauli_op)

    lib.stabilizer_get_pauli_op.restype = ctypes.c_int
    lib.stabilizer_get_pauli_op.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                            ctypes.c_int, ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_get_pauli_op(ctypes.byref(sb), ctypes.c_int(gene_id),
                                      ctypes.c_int(qubit_id), ctypes.byref(c_pauli_op))

    if ret == cfg.FALSE:
        raise ValueError("can't get pauli operator from the Stabilizer object.")

    pauli_op = c_pauli_op.value

    return pauli_op

def stabilizer_operate_qgate(sb, kind=None, q0=None, q1=None):
    """ operate quantum gate to the stabilizer state """

    lib.stabilizer_operate_qgate.restype = ctypes.c_int
    lib.stabilizer_operate_qgate.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                             ctypes.c_int, ctypes.c_int]
    ret = lib.stabilizer_operate_qgate(ctypes.byref(sb), ctypes.c_int(kind),
                                       ctypes.c_int(q0), ctypes.c_int(q1))

    if ret == cfg.FALSE:
        raise ValueError("can't operate quantum gate to the Stabilizer object.")

def stabilizer_get_rank(sb):
    """ get rank of the stabilizer state """

    rank = 0
    c_rank = ctypes.c_int(rank)

    lib.stabilizer_get_rank.restype = ctypes.c_int
    lib.stabilizer_get_rank.argtypes = [ctypes.POINTER(Stabilizer),
                                        ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_get_rank(ctypes.byref(sb), ctypes.byref(c_rank))

    if ret == cfg.FALSE:
        raise ValueError("can't get rank of the Stabilizer object.")

    rank = c_rank.value

    return rank

def stabilizer_measure(sb, q=None):
    """ measure the stabilizer state """

    prob = [0.0, 0.0]
    DoubleArray = ctypes.c_double * 2
    c_prob = DoubleArray(*prob)

    mval = 0
    c_mval = ctypes.c_int(mval)

    lib.stabilizer_measure.restype = ctypes.c_int
    lib.stabilizer_measure.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                       DoubleArray, ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_measure(ctypes.byref(sb), ctypes.c_int(q),
                                 c_prob, ctypes.byref(c_mval))

    if ret == cfg.FALSE:
        raise ValueError("can't measure the Stabilizer object.")

    mval = c_mval.value

    return mval

def stabilizer_operate_qcirc(sb, cmem, qcirc, shots, cid):
    """ operate quantum circuit to the stabilizer state """

    lib.stabilizer_operate_qcirc.restype = ctypes.c_int
    lib.stabilizer_operate_qcirc.argtypes = [ctypes.POINTER(Stabilizer),
                                             ctypes.POINTER(CMem), ctypes.POINTER(QCirc)]

    if cmem is not None:

        cmem_num = cmem.cmem_num
        frequency = Counter()
        for n in range(shots):

            if n < shots - 1:
                sb_tmp = sb.clone()
                cmem_tmp = cmem.clone()
                ret = lib.stabilizer_operate_qcirc(ctypes.byref(sb_tmp),
                                                   ctypes.byref(cmem_tmp), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem_tmp.bit_array, ctypes.POINTER(ctypes.c_ubyte*cmem_num))
            else:
                ret = lib.stabilizer_operate_qcirc(ctypes.byref(sb),
                                                   ctypes.byref(cmem), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem.bit_array, ctypes.POINTER(ctypes.c_ubyte*cmem_num))

            if ret == cfg.FALSE:
                raise ValueError("can't operate quantum circuit to the Stabilizer object.")

            cmem_list = [bit_array.contents[i] for i in range(cmem_num)]
            mval = "".join(map(str, [cmem_list[c] for c in cid]))

            frequency[mval] += 1

        return frequency

    c_cmem = ctypes.POINTER(CMem)()
    ret = lib.stabilizer_operate_qcirc(ctypes.byref(sb), c_cmem, ctypes.byref(qcirc))

    if ret == cfg.FALSE:
        raise ValueError("can't operate quantum circuit to the Stabilizer object.")

    return None

def stabilizer_free(stab):
    """ free memory of Stabilizer object """

    lib.stabilizer_free.argtypes = [ctypes.POINTER(Stabilizer)]
    lib.stabilizer_free(ctypes.byref(stab))



# -*- coding: utf-8 -*-
""" wrapper functions for CMem """
import ctypes
from ctypes.util import find_library
import pathlib
import numpy as np

import qlazy.config as cfg
from qlazy.util import get_lib_ext
from qlazy.CMem import CMem

lib = ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"), mode=ctypes.RTLD_GLOBAL)

def cmem_init(cmem_num):
    """ initialize classical memory """

    if cmem_num < 1:
        raise ValueError("cmem size must be positive integer.")

    cmem = None
    c_cmem = ctypes.c_void_p(cmem)

    lib.cmem_init.restype = ctypes.c_int
    lib.cmem_init.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.cmem_init(ctypes.c_int(cmem_num), c_cmem)

    if ret == cfg.FALSE:
        raise ValueError("can't initialize cmem.")

    return c_cmem

def cmem_copy(cm):
    """ copy classical memory """

    cmem = None
    c_cmem = ctypes.c_void_p(cmem)

    lib.cmem_copy.restype = ctypes.c_int
    lib.cmem_copy.argtypes = [ctypes.POINTER(CMem),
                              ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.cmem_copy(ctypes.byref(cm), c_cmem)

    if ret == cfg.FALSE:
        raise ValueError("can't copy cmem.")

    return c_cmem

def cmem_get_bits(cmem):
    """ get bits of the classical memory """

    if cmem is None:
        raise Valueerror("cmem must be set.")

    cmem_num = cmem.cmem_num
    bits = None
    c_bits = ctypes.c_void_p(bits)

    lib.cmem_get_bits.restype = ctypes.c_int
    lib.cmem_get_bits.argtypes = [ctypes.POINTER(CMem), ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.cmem_get_bits(ctypes.byref(cmem), c_bits)

    if ret == cfg.FALSE:
        raise ValueError("can't get element of the quantum state vector.")

    o = ctypes.cast(c_bits.value, ctypes.POINTER(ctypes.c_ubyte))

    out = [o[i] for i in range(cmem_num)]
    
    libc.free.argtypes = [ctypes.POINTER(ctypes.c_ubyte)]
    libc.free(o)

    return np.array(out)

def cmem_free(cmem):
    """ free classical memory """

    lib.cmem_free.argtypes = [ctypes.POINTER(CMem)]
    lib.cmem_free(ctypes.byref(cmem))



# -*- coding: utf-8 -*-
""" run function for AWS """
from collections import Counter
import numpy as np

from braket.aws import AwsDevice
from braket.devices import LocalSimulator
from braket.circuits import Circuit

import qlazy.config as cfg
from qlazy.Result import Result

def __braket_add_qgate(qc_braket, kind, q0, q1, angle, product):
    """ add gate to braket circuit """

    # 1-qubit, 0-parameter gate
    if kind == cfg.PAULI_X:
        qc_braket.x(q0)

    elif kind == cfg.PAULI_Z:
        qc_braket.z(q0)

    elif kind == cfg.HADAMARD:
        qc_braket.h(q0)

    elif kind == cfg.PHASE_SHIFT_S:
        qc_braket.s(q0)

    elif kind == cfg.PHASE_SHIFT_S_:
        qc_braket.si(q0)

    elif kind == cfg.PHASE_SHIFT_T:
        qc_braket.t(q0)

    elif kind == cfg.PHASE_SHIFT_T_:
        qc_braket.ti(q0)

    # 1-qubit, 1-parameter gate
    elif kind == cfg.ROTATION_X:
        qc_braket.rx(q0, angle=angle)

    elif kind == cfg.ROTATION_Z:
        qc_braket.rz(q0, angle=angle)

    # 2-qubit, 0-parameters gate
    elif kind == cfg.CONTROLLED_X:
        qc_braket.cnot(q0, q1)

    elif kind == cfg.CONTROLLED_Z:
        if product == 'braket_ionq':
            qc_braket.h(q1).cnot(q0, q1).h(q1)
        else:
            qc_braket.cz(q0, q1)

    elif kind == cfg.CONTROLLED_H:
        pi_2 = 0.5 * np.pi
        pi_4 = 0.25 * np.pi
        qc_braket.ry(1, angle=-pi_4).cnot(0, 1).rz(1, angle=-pi_2)
        qc_braket.cnot(0, 1).rz(1, angle=pi_2).ry(1, angle=pi_4)
        if product == 'braket_ionq':
            qc_braket.rz(0, angle=pi_2)
        else:
            qc_braket.phaseshift(0, angle=pi_2)

    # 2-qubit, 1-parameters gate
    elif kind == cfg.CONTROLLED_RZ:
        qc_braket.rz(q1, angle=angle/2).cnot(q0, q1).rz(q1, angle=-angle/2).cnot(q0, q1)

    else:
        raise ValueError("unknown gate")

def __convert_to_braket_circuit(qc_qlazy, product):
    """ convert qlazy circuit to braket circuit """

    # not supported if any unitary gate exist after non-unitary gate (measure,rest)
    qubit_num = qc_qlazy.qubit_num
    measured_info = [-1] * qubit_num # all -1 means no qubits have measured
    qc_qlazy_tmp = qc_qlazy.clone()
    while True:
        kind = qc_qlazy_tmp.kind_first()
        if kind is None:
            break

        (kind, qid, para, c, ctrl) = qc_qlazy_tmp.pop_gate()

        # qid[0]th-qubit has measured and store c-th classical register
        if kind in (cfg.MEASURE, cfg.RESET):
            measured_info[qid[0]] = c
        # not allowed operating unitary gate on measured qubit (braket limitation?)
        elif measured_info[qid[0]] != -1  or measured_info[qid[1]] != -1:
            raise ValueError("not allowed operating unitary gate on measured qubit.")
        elif ctrl is not None:
            raise ValueError("not allowed operating unitary gate on measured qubit.")
        # ex) mesure(qid=[0,1,2], cid=[2,1,0]) -> measured_flg = [2, 1, 0, -1, -1,...]

    # qlazy gate -> braket gate
    qc_braket = Circuit()
    qc_qlazy_tmp = qc_qlazy.clone()
    while True:
        kind = qc_qlazy_tmp.kind_first()
        if kind is None:
            break

        (kind, qid, para, c, ctrl) = qc_qlazy_tmp.pop_gate()

        if kind in (cfg.MEASURE, cfg.RESET):
            continue

        angle = para[0] * np.pi
        q0, q1 = qid[0], qid[1]
        __braket_add_qgate(qc_braket, kind, q0, q1, angle, product)

    return qc_braket, measured_info

def run(qcirc=None, shots=1, cid=None, backend=None):
    """ run the quantum circuit on braket_sv """

    if qcirc is None:
        raise ValueError("quantum circuit must be specified.")

    qubit_num = qcirc.qubit_num
    cmem_num = qcirc.cmem_num

    qc, measured_info = __convert_to_braket_circuit(qcirc, backend.product)

    if backend.product == 'braket_local':
        if backend.device == 'braket_sv':
            device = LocalSimulator(backend="braket_sv")
        else:
            raise ValueError("device:'{}' is unknown for product:'{}'."
                             .format(backend.device, backend.product))
        task = device.run(qc, shots=shots)

    elif backend.product == 'braket_aws':
        if backend.device == 'sv1':
            device = AwsDevice("arn:aws:braket:::device/quantum-simulator/amazon/sv1")
            s3_folder = (backend.config_braket['backet_name'], "sv1")
        elif backend.device == 'tn1':
            device = AwsDevice("arn:aws:braket:::device/quantum-simulator/amazon/tn1")
            s3_folder = (backend.config_braket['backet_name'], "tn1")
        elif backend.device == 'dm1':
            device = AwsDevice("arn:aws:braket:::device/quantum-simulator/amazon/dm1")
            s3_folder = (backend.config_braket['backet_name'], "dm1")
        else:
            raise ValueError("device:'{}' is unknown for product:'{}'."
                             .format(backend.device, backend.product))
        task = device.run(qc, s3_folder, shots=shots)

    elif backend.product == 'braket_ionq':
        if backend.device == 'ionq':
            device = AwsDevice("arn:aws:braket:::device/qpu/ionq/ionQdevice")
            s3_folder = (backend.config_braket['backet_name'], "ionq")
        else:
            raise ValueError("device:'{}' is unknown for product:'{}'."
                             .format(backend.device, backend.product))
        task = device.run(qc, s3_folder, shots=shots)

    elif backend.product == 'braket_rigetti':
        if backend.device == 'aspen_11':
            device = AwsDevice("arn:aws:braket:::device/qpu/rigetti/Aspen-11")
            s3_folder = (backend.config_braket['backet_name'], "aspen_11")
        if backend.device == 'aspen_m_1':
            device = AwsDevice("arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-1")
            s3_folder = (backend.config_braket['backet_name'], "aspen_m_1")
        else:
            raise ValueError("device:'{}' is unknown for product:'{}'."
                             .format(backend.device, backend.product))
        task = device.run(qc, s3_folder, shots=shots)

    elif backend.product == 'braket_oqc':
        if backend.device == 'lucy':
            device = AwsDevice("arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy")
            s3_folder = (backend.config_braket['backet_name'], "lucy")
        else:
            raise ValueError("device:'{}' is unknown for product:'{}'."
                             .format(backend.device, backend.product))
        task = device.run(qc, s3_folder, shots=shots)

    result = task.result()
    frequency_org = result.measurement_counts

    # redefine the measured info
    # ex) mesure(qid=[0,1,2], cid=[2,1,0]) -> measured_info = [2, 1, 0, -1, -1,...]
    #     cid = [1,2] -> measured_info = [1, 0, -1, -1,...]
    #     cid = [2,0] -> measured_info = [0, -1, 1, -1, -1,...]
    #     cid = None -> cid = [0,1,2] -> measured_info = [2, 1, 0, -1, -1,...]

    if cid is None:
        cid = list(range(cmem_num))

    if cmem_num < len(cid):
        raise ValueError("length of cid must be less than classical resister size of qcirc")

    for i, m in enumerate(measured_info):
        if m == -1:
            continue

        if m in cid:
            measured_info[i] = cid.index(m)
        else:
            measured_info[i] = -1

    # marginal frequency
    if set(measured_info) == {-1}:  # no qubits is measured
        frequency = None
    else:
        frequency = Counter()
        for mstr, freq in frequency_org.items():
            mlist = list(mstr)
            mlist_new = ['0'] * len(cid)
            for i, m in enumerate(measured_info):
                if m >= 0:
                    mlist_new[m] = mlist[i]
            mstr_new = "".join(mlist_new)
            frequency[mstr_new] += freq

    info = {'braket': result}

    result = Result()
    result.backend = backend
    result.qubit_num = qubit_num
    result.cmem_num = cmem_num
    result.cid = cid
    result.shots = shots
    result.frequency = frequency
    result.info = info

    return result



# -*- coding: utf-8 -*-
""" run function for qlazy's qstate simulator """

from qlazy.QState import QState
from qlazy.CMem import CMem
from qlazy.Result import Result
from qlazy.lib.qstate_c import qstate_operate_qcirc

def run(qcirc=None, shots=1, cid=None, backend=None):
    """ run the quantum circuit """

    if qcirc is None:
        raise ValueError("quantum circuit must be specified.")

    qubit_num = qcirc.qubit_num
    cmem_num = qcirc.cmem_num
    qstate = QState(qubit_num=qubit_num)
    if cmem_num > 0:
        cmem = CMem(cmem_num)
    else:
        cmem = None

    if cid is None:
        cid = list(range(cmem_num))

    if cmem_num < len(cid):
        raise ValueError("length of cid must be less than classical resister size of qcirc")

    qcirc_unitary, qcirc_non_unitary = qcirc.split_unitary_non_unitary()

    frequency = qstate_operate_qcirc(qstate, cmem, qcirc_unitary, 1, cid)
    frequency = qstate_operate_qcirc(qstate, cmem, qcirc_non_unitary, shots, cid)

    info = {'qstate': qstate, 'cmem': cmem}

    result = Result()
    result.backend = backend
    result.qubit_num = qubit_num
    result.cmem_num = cmem_num
    result.cid = cid
    result.shots = shots
    result.frequency = frequency
    result.info = info

    return result



# -*- coding: utf-8 -*-
""" Backend device of quantum computing """

import os
import errno
import datetime
import configparser

BACKEND_DEVICES = {'qlazy': ['qstate_simulator', 'stabilizer_simulator'],
                   'qulacs': ['cpu_simulator', 'gpu_simulator'],
                   'ibmq': ['aer_simulator'],
                   'braket_local': ['braket_sv'],
                   'braket_aws': ['sv1', 'tn1', 'dm1'],
                   'braket_ionq': ['ionq'],
                   'braket_rigetti': ['aspen_11', 'aspen_m_1'],
                   'braket_oqc': ['lucy']}

class Backend:
    """ Backend device of quantum computing

    Attributes
    ----------
    product : str
        product name of the backend
        ('qlazy' or 'qulacs' or 'ibmq')
    device : str
        device name of the product

    Notes
    -----
    If you use amazon braket backend (braket_local, braket_aws,
    braket_ionq, braket_rigetti, braket_oqc), you must have
    config.ini file in your '~/.qlazy' directory,
    and discribe like following example ...

    Example
    -------
    >>> bk = Backend(product='qlazy', device='stabilizer_simulator')
    >>> bk = Backend(product='qulacs', device='gpu_simulator')
    >>> bk = Backend(product='ibmq', device='least_busy')
    >>> bk = Backend(product='braket_rigetti', device='aspen_11')
    ...
    $ cat ~/.qlazy/config.ini
    [backend_braket]
    backet_name = amazon-braket-your-S3-baket-name
    ...
    $ cat ~/.qlazy/config.ini
    [backend_braket]
    backet_name = amazon-braket-your-S3-baket-name
    poll_timeout_seconds = 86400  # set 1-day (default: 5-days)

    """
    def __init__(self, product=None, device=None):

        #
        # config
        #

        config_ini = configparser.ConfigParser()
        config_ini_path = os.environ['HOME'] + '/.qlazy/config.ini'
        if not os.path.exists(config_ini_path):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), config_ini_path)
        config_ini.read(config_ini_path, encoding='utf-8')

        # braket
        backend_braket = config_ini['backend_braket']
        backet_name = backend_braket.get('backet_name')
        poll_timeout_seconds = backend_braket.get('poll_timeout_seconds')
        self.config_braket = {'backet_name': backet_name,
                              'poll_timeout_seconds': poll_timeout_seconds}

        #
        # set attributes (procuct, device)
        #

        if product is None:
            self.product = 'qlazy'
            if device is None:
                self.device = 'qstate_simulator'
            elif device in BACKEND_DEVICES[product]:
                self.device = device
            else:
                raise ValueError("device:'{}' is unknown for product:'{}'."
                                 .format(device, self.product))

        elif product in BACKEND_DEVICES.keys():
            self.product = product
            if device in BACKEND_DEVICES[product]:
                self.device = device
            elif device is None:
                self.device = BACKEND_DEVICES[product][0]
            elif product == 'ibmq':
                self.device = device
            else:
                raise ValueError("device:'{}' is unknown for product:'{}'."
                                 .format(device, self.product))

        else:
            raise ValueError("product:{} is unknown.".format(product))

        #
        # set method (__run)
        #

        # qlazy
        if self.product == 'qlazy':

            # qstate simulator
            if self.device == 'qstate_simulator':
                from qlazy.backend.qlazy_qstate_simulator import run
                self.__run = run

            # stabilizer simulator
            elif self.device == 'stabilizer_simulator':
                from qlazy.backend.qlazy_stabilizer_simulator import run
                self.__run = run

            else:
                raise ValueError("device:'{}' is unknown for product:'{}'."
                                 .format(self.device, self.product))

        # qulacs
        elif self.product == 'qulacs':

            # cpu_simulator
            if self.device == 'cpu_simulator':
                from qlazy.backend.qulacs_simulator import run_cpu
                self.__run = run_cpu

            # gpu_simulator
            elif self.device == 'gpu_simulator':
                from qlazy.backend.qulacs_simulator import run_gpu
                self.__run = run_gpu

            else:
                raise ValueError("device:'{}' is unknown for product:'{}'."
                                 .format(self.device, self.product))

        # ibmq
        elif self.product == 'ibmq':
            from qlazy.backend.ibmq import run
            self.__run = run

        # braket
        elif self.product in ('braket_local', 'braket_aws',
                              'braket_ionq', 'braket_rigetti', 'braket_oqc'):
            from qlazy.backend.amazon_braket import run
            self.__run = run

        else:
            raise ValueError("device:'{}' is unknown for product:'{}'."
                             .format(self.device, self.product))

    @classmethod
    def products(cls):
        """ get products list """
        return list(BACKEND_DEVICES)

    @classmethod
    def ibmq_devices(cls):
        """ get ibmq's devices list """
        devices = BACKEND_DEVICES['ibmq']
        try:
            from qiskit import IBMQ
            provider = IBMQ.load_account()
            ibmq_backend_system_names = [b.name() for b in
                                         provider.backends(simulator=False, operational=True)]
        except Exception:
            pass
        else:
            devices += ['least_busy']
            devices += ibmq_backend_system_names

        return devices

    @classmethod
    def devices(cls, product):
        """ get devices list for the product """

        if product in ('qlazy', 'qulacs', 'braket_local', 'braket_aws',
                       'braket_ionq', 'braket_rigetti', 'braket_oqc'):
            device_list = BACKEND_DEVICES[product]
        elif product == 'ibmq':
            device_list = Backend.ibmq_devices()
        else:
            raise ValueError("unknown product: {}".format(product))

        return device_list

    def __str__(self):

        backend_dict = {'product': self.product, 'device': self.device}
        return str(backend_dict)

    def run(self, qcirc=None, shots=1, cid=None):
        """
        run the quantum circuit.

        Parameters
        ----------
        qcirc : instance of QCirc, default None
            quantum circuit.
        shots : int, default 1
            number of measurements.
        cid : list, default []
            classical register id list to count frequency.

        Returns
        -------
        result : incetance of Result
            measurement result.

        Examples
        --------
        >>> # simple example
        >>> from qlazy import QCirc, Backend
        >>> bk = Backend(product='qlazy', device='qstate_simulator')
        >>> qc = QCirc().h(0).cx(0,1).measure(qid=[0,1], cid=[0,1])
        >>> result = bk.run(qcirc=qc, shots=100)
        >>> print(result.frequency)
        Counter({'00': 51, '11': 49})
        >>> ...
        >>> # control quantum gate by measured results
        >>> from qlazy import QCirc, Backend
        >>> bk = Backend(product='qlazy', device='qstate_simulator')
        >>> qc = QCirc.h(0)
        >>> qc.cx(0,1).measure(qid=[0],cid=[0])
        >>> qc.x(0, ctrl=0).x(1, ctrl=0).measure(qid=[0,1], cid=[0,1])
        >>> result = bk.run(qcirc=qc, shots=100)
	>>> print(result.frequency)
        Counter({'00': 100})

        """
        start_time = datetime.datetime.now()
        result = self.__run(qcirc=qcirc, shots=shots, cid=cid, backend=self)
        end_time = datetime.datetime.now()

        result.start_time = start_time
        result.end_time = end_time
        result.elapsed_time = (end_time - start_time).total_seconds()

        return result



# -*- coding: utf-8 -*-
""" multi-controlled cnot for density operator """

def __gray_code(n):
    """ gray code generator """

    for k in range(2**n):
        yield k^(k>>1)

def densop_mcx(de, qid=None):
    """ multi-controlled cnot """

    if qid is None:
        raise ValueError("qid must be set.")

    # controled and target register
    qid_ctr = qid[:-1]
    qid_tar = qid[-1]

    # hadamard
    de.h(qid_tar)

    # controlled-RZ(psi), psi=pi/(2**(bitnum-1))
    bitnum = len(qid_ctr)
    psi = 1.0/(2**(bitnum-1)) # unit=pi(radian)
    gray_pre = 0
    for gray in __gray_code(bitnum):
        if gray == 0:
            continue
        msb = len(str(bin(gray)))-3
        chb = len(str(bin(gray^gray_pre)))-3
        if gray != 1:
            if chb == msb:
                chb -= 1
            de.cx(qid_ctr[chb], qid_ctr[msb])
        de.cp(qid_ctr[msb], qid_tar, phase=psi)
        psi = -psi
        gray_pre = gray

    # hadamard
    de.h(qid_tar)



# -*- coding: utf-8 -*-
""" Result of quantum circuit execution """

from collections import Counter
import datetime
import pickle

from qlazy.Backend import Backend

class Result:
    """ Result of quantum circuit execution

    Attributes
    ----------
    backend : instance of Backend
        backend device of quantum computing
    qubit_num : int
        number of qubits
    cmem_num : int
        number of classical bits
    cid : list
        classical register id list to get frequencys.
    shots : int
        number of measurements
    freqency : instance of Counter
        frequencies of measured value.
    start_time : instance of datetime.datetime
        start time to ececute the quantum circuit.
    end_time : instance of datetime.datetime
        end time to ececute the quantum circuit.
    elapsed_time : float
        elapsed time to ececute the quantum circuit.
    info : dict
        result informations relating to the backend device

    """
    def __init__(self):

        self.__backend = None
        self.__qubit_num = None
        self.__cmem_num = None
        self.__cid = None
        self.__shots = None
        self.__frequency = None
        self.__start_time = None
        self.__end_time = None
        self.__elapsed_time = None
        self.__info = None

    # backend

    @property
    def backend(self):
        """ getter of backend """
        return self.__backend

    @backend.setter
    def backend(self, backend):
        """ setter of backend """
        if isinstance(backend, Backend) is False:
            raise TypeError("type of backend must be Backend.")
        self.__backend = backend

    # qubit_num

    @property
    def qubit_num(self):
        """ getter of qubit_num """
        return self.__qubit_num

    @qubit_num.setter
    def qubit_num(self, qubit_num):
        """ setter of qubit_num """
        if isinstance(qubit_num, int) is False:
            raise TypeError("type of qubit_num must be int.")
        self.__qubit_num = qubit_num

    # cmem_num

    @property
    def cmem_num(self):
        """ getter of cmem_num """
        return self.__cmem_num

    @cmem_num.setter
    def cmem_num(self, cmem_num):
        """ getter of cmem_nem """
        if isinstance(cmem_num, int) is False:
            raise TypeError("type of cmem_num must be int.")
        self.__cmem_num = cmem_num

    # cid

    @property
    def cid(self):
        """ getter of cid """
        return self.__cid

    @cid.setter
    def cid(self, cid):
        """ setter of cid """
        if isinstance(cid, list) is False:
            raise TypeError("type of cid must be list.")
        self.__cid = cid

    # shots

    @property
    def shots(self):
        """ getter of shots """
        return self.__shots

    @shots.setter
    def shots(self, shots):
        """ setter of shots """
        if isinstance(shots, int) is False:
            raise TypeError("type of shots must be int.")
        self.__shots = shots

    # frequency

    @property
    def frequency(self):
        """ getter of frequency """
        return self.__frequency

    @frequency.setter
    def frequency(self, frequency):
        """ setter of frequency """
        if isinstance(frequency, Counter) is False and frequency is not None:
            raise TypeError("type of frequency must be Counter.")
        self.__frequency = frequency

    # start_time

    @property
    def start_time(self):
        """ getter of start_time """
        return self.__start_time

    @start_time.setter
    def start_time(self, start_time):
        """ setter of start_time """
        if isinstance(start_time, datetime.datetime) is False:
            raise TypeError("type of sart_time must be datetime.datetime.")
        self.__start_time = start_time

    # end_time

    @property
    def end_time(self):
        """ getter of end_time """
        return self.__end_time

    @end_time.setter
    def end_time(self, end_time):
        """ setter of end_time """
        if isinstance(end_time, datetime.datetime) is False:
            raise TypeError("type of end_time must be datetime.datetime.")
        self.__end_time = end_time

    # elapsed_time

    @property
    def elapsed_time(self):
        """ getter of elapsed_time """
        return self.__elapsed_time

    @elapsed_time.setter
    def elapsed_time(self, elapsed_time):
        """ setter of elapsed_time """
        if isinstance(elapsed_time, float) is False:
            raise TypeError("type of elaased_time must be float.")
        self.__elapsed_time = elapsed_time

    # info

    @property
    def info(self):
        """ getter of info
            (for device dependent informattion)
        """
        return self.__info

    @info.setter
    def info(self, info):
        """ setter of info """
        if isinstance(info, dict) is False and info is not None:
            raise TypeError("type of info must be dict.")
        self.__info = info

    def __str__(self):

        s = "backend: {}\n".format(self.__backend)
        s += "qubit_num: {}\n".format(self.__qubit_num)
        s += "cmem_num: {}\n".format(self.__cmem_num)
        s += "cid: {}\n".format(self.__cid)
        s += "shots: {}\n".format(self.__shots)
        s += "frequency: {}\n".format(self.__frequency)
        s += "start_time: {}\n".format(self.__start_time)
        s += "end_time: {}\n".format(self.__end_time)
        s += "elapsed_time: {}\n".format(self.__elapsed_time)
        s += "info: {}\n".format(self.__info)

        return s

    def show(self, verbose=False):
        """
        show the result.

        Parameters
        ----------
        verbose: bool
            verbose output or not
        None

        Returns
        -------
        None

        """
        s = ""
        if verbose is True:
            s += "[backend]\n"
            s += "- product      = {}\n".format(self.__backend.product)
            s += "- device       = {}\n".format(self.__backend.device)
            s += "[qubit & cmem]\n"
            s += "- qubit_num    = {}\n".format(self.__qubit_num)
            s += "- cmem_num     = {}\n".format(self.__cmem_num)
            s += "[measurement]\n"
            s += "- cid          = {}\n".format(self.__cid)
            s += "- shots        = {}\n".format(self.__shots)
            s += "[time]\n"
            s += "- start_time   = {}\n".format(self.__start_time)
            s += "- end_time     = {}\n".format(self.__end_time)
            s += "- elapsed_time = {:.6f} [sec]\n".format(self.__elapsed_time)
            s += "[histogram]\n"

        if self.__frequency is None:
            s += "None\n"
        else:
            digits = len(str(max(self.__frequency.values())))

            for k, v in self.__frequency.items():
                prob = v / self.__shots
                if v > 0:
                    bar_graph = '+'
                bar_graph += '+' * int(prob * 30)
                if verbose is True:
                    bp = "- "
                else:
                    bp = ""
                s += (bp + "freq[{0:}] = {1:{digits}d} ({2:1.4f}) |{3:}\n"
                      .format(k, v, prob, bar_graph, digits=digits))

        print(s.rstrip())

    def save(self, file_path):
        """
        save the result

        Parameters
        ----------
        file_path: str
            file path of dump file

        Returns
        -------
        None

        """
        result_dict = {'backend': self.__backend,
                       'qubit_num': self.__qubit_num, 'cmem_num': self.__cmem_num,
                       'cid': self.__cid, 'shots': self.__shots,
                       'frequency': self.__frequency, 'start_time': self.__start_time,
                       'end_time': self.__end_time, 'elapsed_time': self.__elapsed_time}

        with open(file_path, mode='wb') as f:
            pickle.dump(result_dict, f)

    @classmethod
    def load(cls, file_path):
        """
        load the result

        Parameters
        ----------
        file_path: str
            file path of dump file

        Returns
        -------
        result: instance of Result
            loaded circuit

        """
        with open(file_path, mode='rb') as f:
            result_dict = pickle.load(f)

        result = Result()
        result.backend = result_dict['backend']
        result.qubit_num = result_dict['qubit_num']
        result.cmem_num = result_dict['cmem_num']
        result.cid = result_dict['cid']
        result.shots = result_dict['shots']
        result.frequency = result_dict['frequency']
        result.start_time = result_dict['start_time']
        result.end_time = result_dict['end_time']
        result.elapsed_time = result_dict['elapsed_time']

        return result



# -*- coding: utf-8 -*-
""" various kind of utilities """
import math
import numpy as np

import qlazy.config as cfg

def get_lib_ext():
    """ get library extension (for Mac OS)"""

    import platform
    if platform.system() == 'Darwin':
        return 'dylib'
    return 'so'

def qstate_check_args(qs, kind=None, qid=None):
    """ check arguments for qstate """

    for q in qid:
        if (q >= qs.qubit_num) or (q < 0):
            raise IndexError("index out of range.")

    qnum = get_qgate_qubit_num(kind)

    if qnum == 0:  # any qubit number
        # check qubit number
        if len(qid) > qs.qubit_num:
            raise IndexError("Too many arguments.")
        if len(qid) < 1:
            raise IndexError("Need more arguments.")
        # check same qubit number
        if len(set(qid)) != len(qid):
            raise IndexError("Same qubit ID.")

    elif qnum == 1:
        # check qubit number
        if len(qid) > qnum:
            raise IndexError("Too many arguments.")
        if len(qid) < qnum:
            raise IndexError("Need more arguments.")

    elif qnum == 2:
        # check qubit number
        if len(qid) > qnum:
            raise IndexError("Too many arguments.")
        if len(qid) < qnum:
            raise IndexError("Need more arguments.")
        # check same qubit number
        if qid[0] == qid[1]:
            raise IndexError("Same qubit ID.")

    elif qnum == 3:
        # check qubit number
        if len(qid) > qnum:
            raise IndexError("Too many arguments.")
        if len(qid) < qnum:
            raise IndexError("Need more arguments.")
        # check same qubit id
        if qid[0] == qid[1] or qid[1] == qid[2] or qid[2] == qid[0]:
            raise IndexError("Same qubit ID.")

    return True

def densop_check_args(de, kind=None, qid=None):
    """ check arguments for densop """

    qubit_num = int(math.log2(de.row))

    for q in qid:
        if (q >= qubit_num) or (q < 0):
            raise IndexError("index out of range.")

    qnum = get_qgate_qubit_num(kind)

    if qnum == 0:  # any qubit number
        # check qubit number
        if len(qid) > qubit_num:
            raise IndexError("Too many arguments.")
        if len(qid) < 1:
            raise IndexError("Need more arguments.")

        # check same qubit number
        if len(set(qid)) != len(qid):
            raise IndexError("Same qubit ID.")

    elif qnum == 1:
        # check qubit number
        if len(qid) > qnum:
            raise IndexError("Too many arguments.")
        if len(qid) < qnum:
            raise IndexError("Need more arguments.")

    elif qnum == 2:
        # check qubit number
        if len(qid) > qnum:
            raise IndexError("Too many arguments.")
        if len(qid) < qnum:
            raise IndexError("Need more arguments.")

        # check same qubit number
        if qid[0] == qid[1]:
            raise IndexError("Same qubit ID.")

    elif qnum == 3:
        # check qubit number
        if len(qid) > qnum:
            raise IndexError("Too many arguments.")
        if len(qid) < qnum:
            raise IndexError("Need more arguments.")

        # check same qubit id
        if qid[0] == qid[1] or qid[1] == qid[2] or qid[2] == qid[0]:
            raise IndexError("Same qubit ID.")

    return True

def get_qgate_qubit_num(kind=None):
    """ get qubit number for the quantum gate """

    if kind in (cfg.SHOW, cfg.MEASURE, cfg.MEASURE_X, cfg.MEASURE_Y,
                cfg.MEASURE_Z, cfg.RESET):  # 0 if any number
        return 0
    if ((kind in (cfg.BLOCH, cfg.PAULI_X, cfg.PAULI_Y, cfg.PAULI_Z, cfg.ROOT_PAULI_X,
                  cfg.ROOT_PAULI_X_, cfg.HADAMARD, cfg.PHASE_SHIFT_S, cfg.PHASE_SHIFT_S_,
                  cfg.PHASE_SHIFT_T, cfg.PHASE_SHIFT_T_, cfg.PHASE_SHIFT, cfg.ROTATION_X,
                  cfg.ROTATION_Y, cfg.ROTATION_Z, cfg.ROTATION_U1, cfg.ROTATION_U2,
                  cfg.ROTATION_U3))):
        return 1
    if (kind in (cfg.CONTROLLED_X, cfg.CONTROLLED_Y, cfg.CONTROLLED_Z, cfg.CONTROLLED_XR,
                 cfg.CONTROLLED_XR_, cfg.CONTROLLED_H, cfg.CONTROLLED_S, cfg.CONTROLLED_S_,
                 cfg.CONTROLLED_T, cfg.CONTROLLED_T_, cfg.SWAP_QUBITS, cfg.CONTROLLED_P,
                 cfg.CONTROLLED_RX, cfg.CONTROLLED_RY, cfg.CONTROLLED_RZ, cfg.CONTROLLED_U1,
                 cfg.CONTROLLED_U2, cfg.CONTROLLED_U3, cfg.MEASURE_BELL,
                 cfg.ROTATION_XX, cfg.ROTATION_YY, cfg.ROTATION_ZZ)):
        return 2

    raise ValueError("unknown quantum gate.")

def get_qgate_param_num(kind=None):
    """ get parameter number for the quantum gate """

    if (kind in (cfg.SHOW, cfg.BLOCH, cfg.MEASURE, cfg.MEASURE_X, cfg.MEASURE_Y,
                 cfg.MEASURE_Z, cfg.MEASURE_BELL, cfg.RESET, cfg.PAULI_X, cfg.PAULI_Y,
                 cfg.PAULI_Z, cfg.ROOT_PAULI_X, cfg.ROOT_PAULI_X_, cfg.HADAMARD,
                 cfg.PHASE_SHIFT_S, cfg.PHASE_SHIFT_S_, cfg.PHASE_SHIFT_T,
                 cfg.PHASE_SHIFT_T_, cfg.CONTROLLED_X, cfg.CONTROLLED_Y, cfg.CONTROLLED_Z,
                 cfg.CONTROLLED_XR, cfg.CONTROLLED_XR_, cfg.CONTROLLED_H,
                 cfg.CONTROLLED_S, cfg.CONTROLLED_S_, cfg.CONTROLLED_T, cfg.CONTROLLED_T_,
                 cfg.SWAP_QUBITS)):
        return 0
    if (kind in (cfg.PHASE_SHIFT, cfg.ROTATION_X, cfg.ROTATION_Y, cfg.ROTATION_Z,
                 cfg.ROTATION_U1, cfg.CONTROLLED_P, cfg.CONTROLLED_RX, cfg.CONTROLLED_RY,
                 cfg.CONTROLLED_RZ, cfg.CONTROLLED_U1,
                 cfg.ROTATION_XX, cfg.ROTATION_YY, cfg.ROTATION_ZZ)):
        return 1
    if kind in (cfg.ROTATION_U2, cfg.CONTROLLED_U2):
        return 2
    if kind in (cfg.ROTATION_U3, cfg.CONTROLLED_U3):
        return 3

    raise ValueError("unknown quantum gate.")

def is_clifford_gate(kind):
    """ is the gate clifford? """

    return kind in (cfg.PAULI_X, cfg.PAULI_Y, cfg.PAULI_Z, cfg.HADAMARD,
                    cfg.PHASE_SHIFT_S, cfg.PHASE_SHIFT_S_,
                    cfg.CONTROLLED_X, cfg.CONTROLLED_Y, cfg.CONTROLLED_Z)

def is_non_clifford_gate(kind):
    """ is the gate non-clifford? """

    return (is_clifford_gate(kind) is False and
            is_measurement_gate(kind) is False and
            is_reset_gate(kind) is False)

def is_measurement_gate(kind):
    """ is the gate measurement? """

    return kind in (cfg.MEASURE, cfg.MEASURE_X, cfg.MEASURE_Y,
                    cfg.MEASURE_Z, cfg.MEASURE_BELL)

def is_reset_gate(kind):
    """ is the gate reset? """

    return kind == cfg.RESET

def reverse_bit_order(vec_in):
    """ reverse bit order of the state vector """

    vec_out = [0] * len(vec_in)
    digits = float(np.log2(len(vec_in)))
    if not digits.is_integer():
        raise ValueError("length of vector must be power of two")

    for idx, val in enumerate(vec_in):
        idx_binstr = '{:0{digits}b}'.format(idx, digits=int(digits))
        idx_binstr_rev = ''.join(reversed(list(idx_binstr)))
        idx_rev = int(idx_binstr_rev, 2)
        vec_out[idx_rev] = val

    return vec_out



# -*- coding: utf-8 -*-
""" wrapper functions for QState """
import ctypes
from ctypes.util import find_library
import pathlib
from collections import Counter
import numpy as np

import qlazy.config as cfg
from qlazy.util import get_lib_ext, qstate_check_args
from qlazy.QState import QState
from qlazy.MData import MData
from qlazy.Observable import Observable
from qlazy.QCirc import QCirc
from qlazy.CMem import CMem

lib = ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"), mode=ctypes.RTLD_GLOBAL)

def qstate_init(qubit_num=None, seed=None):
    """ initialize QState object """

    libc.srand(ctypes.c_int(seed))

    qstate = None
    c_qstate = ctypes.c_void_p(qstate)

    lib.qstate_init.restype = ctypes.c_int
    lib.qstate_init.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qstate_init(ctypes.c_int(qubit_num), c_qstate)

    if ret == cfg.FALSE:
        raise ValueError("can't initialize QState object.")

    return c_qstate

def qstate_init_with_vector(vector=None, seed=None):
    """ initialize QState object with vector """

    libc.srand(ctypes.c_int(seed))

    qstate = None
    c_qstate = ctypes.c_void_p(qstate)

    dim = len(vector)
    vec_real = [0.0 for _ in range(dim)]
    vec_imag = [0.0 for _ in range(dim)]
    for i in range(dim):
        vec_real[i] = vector[i].real
        vec_imag[i] = vector[i].imag

    DoubleArray = ctypes.c_double * dim
    c_vec_real = DoubleArray(*vec_real)
    c_vec_imag = DoubleArray(*vec_imag)

    lib.qstate_init_with_vector.restype = ctypes.c_int
    lib.qstate_init_with_vector.argtypes = [DoubleArray, DoubleArray, ctypes.c_int,
                                            ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qstate_init_with_vector(c_vec_real, c_vec_imag, ctypes.c_int(dim),
                                      c_qstate)

    if ret == cfg.FALSE:
        raise ValueError("can't initialize QState object.")

    return c_qstate


def qstate_reset(qs, qid=None):
    """ reset quantum state vector """

    if qid is None or qid == []:
        qid = list(range(qs.qubit_num))

    qstate_check_args(qs, kind=cfg.SHOW, qid=qid)

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i, q in enumerate(qid):
            qubit_id[i] = q

        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        lib.qstate_reset.restype = ctypes.c_int
        lib.qstate_reset.argtypes = [ctypes.POINTER(QState), ctypes.c_int, IntArray]
        ret = lib.qstate_reset(ctypes.byref(qs), ctypes.c_int(qubit_num), qid_array)

        if ret == cfg.FALSE:
            raise ValueError("can't reset quantum state vector.")

    except Exception:
        raise ValueError("can't reset quantum state vector.")


def qstate_print(qs, qid=None, nonzero=False):
    """ print quantum state vector """

    if qid is None or qid == []:
        qid = list(range(qs.qubit_num))

    qstate_check_args(qs, kind=cfg.SHOW, qid=qid)

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i, q in enumerate(qid):
            qubit_id[i] = q

        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        lib.qstate_print.restype = ctypes.c_int
        lib.qstate_print.argtypes = [ctypes.POINTER(QState), ctypes.c_int, IntArray, ctypes.c_bool]
        ret = lib.qstate_print(ctypes.byref(qs), ctypes.c_int(qubit_num),
                               qid_array, ctypes.c_bool(nonzero))

        if ret == cfg.FALSE:
            raise ValueError("can't print quantum state vector.")

    except Exception:
        raise ValueError("can't print quantum state vector.")


def qstate_copy(qs):
    """ copy the quantum state vector """

    try:
        qstate = None
        c_qstate = ctypes.c_void_p(qstate)

        lib.qstate_copy.restype = ctypes.c_int
        lib.qstate_copy.argtypes = [ctypes.POINTER(QState),
                                    ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.qstate_copy(ctypes.byref(qs), c_qstate)

        if ret == cfg.FALSE:
            raise ValueError("can't copy quantum state vector.")

        return c_qstate

    except Exception:
        raise ValueError("can't copy quantum state vector.")


def qstate_bloch(qs, q=0):
    """ get bloch angle from quantum state vector """

    # error check
    qstate_check_args(qs, kind=cfg.BLOCH, qid=[q])

    try:
        theta = 0.0
        phi = 0.0
        c_theta = ctypes.c_double(theta)
        c_phi = ctypes.c_double(phi)

        lib.qstate_bloch.restype = ctypes.c_int
        lib.qstate_bloch.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                     ctypes.POINTER(ctypes.c_double),
                                     ctypes.POINTER(ctypes.c_double)]
        ret = lib.qstate_bloch(ctypes.byref(qs), ctypes.c_int(q),
                               ctypes.byref(c_theta), ctypes.byref(c_phi))

        if ret == cfg.FALSE:
            raise ValueError("can't get bloch angle.")

        theta = c_theta.value
        phi = c_phi.value

        return theta, phi

    except Exception:
        raise ValueError("can't get bloch angle.")


def qstate_inner_product(qs_0, qs_1):
    """ get inner product of 2 quantum state vectors """

    try:
        real = 0.0
        imag = 0.0
        c_real = ctypes.c_double(real)
        c_imag = ctypes.c_double(imag)

        lib.qstate_inner_product.restype = ctypes.c_int
        lib.qstate_inner_product.argtypes = [ctypes.POINTER(QState),
                                             ctypes.POINTER(QState),
                                             ctypes.POINTER(ctypes.c_double),
                                             ctypes.POINTER(ctypes.c_double)]
        ret = lib.qstate_inner_product(ctypes.byref(qs_0), ctypes.byref(qs_1),
                                       ctypes.byref(c_real), ctypes.byref(c_imag))

        if ret == cfg.FALSE:
            raise ValueError("can't get inner product of 2 quantum state vectors.")

        real = c_real.value
        imag = c_imag.value

        return complex(real, imag)

    except Exception:
        raise ValueError("can't get inner product of 2 quantum state vectors.")


def qstate_get_camp(qs, qid=None):
    """ get elements of the quantum state vector """

    if qid is None or qid == []:
        qid = list(range(qs.qubit_num))

    # error check
    if len(qid) > qs.qubit_num:
        raise IndexError("too many arguments.")
    for q in qid:
        if q >= qs.qubit_num:
            raise IndexError("out of range.")
        if q < 0:
            raise IndexError("out of range.")

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i, q in enumerate(qid):
            qubit_id[i] = q
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        camp = None
        c_camp = ctypes.c_void_p(camp)
        lib.qstate_get_camp.restype = ctypes.c_int
        lib.qstate_get_camp.argtypes = [ctypes.POINTER(QState), ctypes.c_int, IntArray,
                                        ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.qstate_get_camp(ctypes.byref(qs), ctypes.c_int(qubit_num),
                                  qid_array, c_camp)

        if ret == cfg.FALSE:
            raise ValueError("can't get element of the quantum state vector.")

        o = ctypes.cast(c_camp.value, ctypes.POINTER(ctypes.c_double))

        state_num = (1 << len(qid))
        out = [0] * state_num
        for i in range(state_num):
            out[i] = complex(o[2*i], o[2*i+1])

        libc.free.argtypes = [ctypes.POINTER(ctypes.c_double)]
        libc.free(o)

    except Exception:
        raise ValueError("can't get element of the quantum state vector.")

    return np.array(out)


def qstate_tensor_product(qs, qstate):
    """ get tensor procuct of 2 quantum state vectors """

    try:
        qstate_out = None
        c_qstate_out = ctypes.c_void_p(qstate_out)

        lib.qstate_tensor_product.restype = ctypes.c_int
        lib.qstate_tensor_product.argtypes = [ctypes.POINTER(QState),
                                              ctypes.POINTER(QState),
                                              ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.qstate_tensor_product(ctypes.byref(qs), ctypes.byref(qstate),
                                        c_qstate_out)

        if ret == cfg.FALSE:
            raise ValueError("can't get tensor product of the 2 quantum state vectors.")

        return c_qstate_out

    except Exception:
        raise ValueError("can't get tensor product of the 2 quantum state vectors.")


def qstate_evolve(qs, observable=None, time=0.0, iteration=0):
    """ time evolution of the quantum state vectors """

    if iteration < 1:
        raise ValueError("iteration must be positive integer.")

    if observable is None:
        raise ValueError("observable must be set.")

    try:
        lib.qstate_evolve.restype = ctypes.c_int
        lib.qstate_evolve.argtypes = [ctypes.POINTER(QState), ctypes.POINTER(Observable),
                                      ctypes.c_double, ctypes.c_int]
        ret = lib.qstate_evolve(ctypes.byref(qs), ctypes.byref(observable),
                                ctypes.c_double(time), ctypes.c_int(iteration))

        if ret == cfg.FALSE:
            raise ValueError("can't get the quantum state vectors after time evolution.")

    except Exception:
        raise ValueError("can't get the quantum state vectors after time evolution.")


def qstate_expect_value(qs, observable=None):
    """ get expectation value of the observable
        under the quantum state vector """

    if observable is None:
        raise ValueError("observable must be set.")

    try:
        val = 0.0
        c_val = ctypes.c_double(val)
        lib.qstate_expect_value.restype = ctypes.c_int
        lib.qstate_expect_value.argtypes = [ctypes.POINTER(QState),
                                            ctypes.POINTER(Observable),
                                            ctypes.POINTER(ctypes.c_double)]
        ret = lib.qstate_expect_value(ctypes.byref(qs),
                                      ctypes.byref(observable),
                                      ctypes.byref(c_val))

        if ret == cfg.FALSE:
            raise ValueError("can't get expect value of the observable"
                             " under the quantum state vector.")

        val = c_val.value

        return complex(val, 0.0)

    except Exception:
        raise ValueError("can't get expect value of the observable"
                         " under the quantum state vector.")

    # return out


def qstate_apply_matrix(qs, matrix=None, qid=None):
    """ apply matrix to the quantum state """

    if matrix is None:
        raise ValueError("matrix must be set.")

    if qid is None or qid == []:
        qid = list(range(qs.qubit_num))

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i, q in enumerate(qid):
            qubit_id[i] = q
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        row = len(matrix) # dimension of the unitary matrix
        col = row
        size = row * col

        mat_complex = []
        for mat_row in matrix:
            for mat_elm in mat_row:
                mat_complex.append(mat_elm)

        mat_real = [0.0 for _ in range(size)]
        mat_imag = [0.0 for _ in range(size)]
        for i in range(size):
            mat_real[i] = mat_complex[i].real
            mat_imag[i] = mat_complex[i].imag

        DoubleArray = ctypes.c_double * size
        c_mat_real = DoubleArray(*mat_real)
        c_mat_imag = DoubleArray(*mat_imag)

        lib.qstate_apply_matrix.restype = ctypes.c_int
        lib.qstate_apply_matrix.argtypes = [ctypes.POINTER(QState),
                                            ctypes.c_int, IntArray,
                                            DoubleArray, DoubleArray,
                                            ctypes.c_int, ctypes.c_int]
        ret = lib.qstate_apply_matrix(ctypes.byref(qs),
                                      ctypes.c_int(qubit_num), qid_array,
                                      c_mat_real, c_mat_imag,
                                      ctypes.c_int(row), ctypes.c_int(col))

        if ret == cfg.FALSE:
            raise ValueError("can't apply the matrix to the quantum state vector.")

    except Exception:
        raise ValueError("can't apply the matrix to the quantum state vector.")


def qstate_operate_qgate(qs, kind=None, qid=None, phase=cfg.DEF_PHASE,
                         phase1=cfg.DEF_PHASE, phase2=cfg.DEF_PHASE):
    """ operate quantum gate to the quantum state """

    # error check
    qstate_check_args(qs, kind=kind, qid=qid)

    qubit_id = [-1 for _ in range(2)]
    for i, q in enumerate(qid):
        qubit_id[i] = q
    IntArray = ctypes.c_int * 2
    qid_array = IntArray(*qubit_id)

    lib.qstate_operate_qgate.restype = ctypes.c_int
    lib.qstate_operate_qgate.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                         ctypes.c_double, ctypes.c_double,
                                         ctypes.c_double, IntArray]
    ret = lib.qstate_operate_qgate(ctypes.byref(qs), ctypes.c_int(kind),
                                   ctypes.c_double(phase), ctypes.c_double(phase1),
                                   ctypes.c_double(phase2), qid_array)

    if ret == cfg.FALSE:
        raise ValueError("can't operate quantum gate to the quantum state vector.")

def qstate_measure(qs, qid=None, angle=0.0, phase=0.0):
    """ measurement of the qubits """

    if qid is None or qid == []:
        qid = list(range(qs.qubit_num))

    # operate
    qubit_num = len(qid)
    qubit_id = [0 for _ in range(qubit_num)]
    for i, q in enumerate(qid):
        qubit_id[i] = q
    IntArray = ctypes.c_int * qubit_num
    qid_array = IntArray(*qubit_id)

    mval = 0
    c_mval = ctypes.c_int(mval)

    lib.qstate_measure.restype = ctypes.c_int
    lib.qstate_measure.argtypes = [ctypes.POINTER(QState),
                                   ctypes.c_double, ctypes.c_double,
                                   ctypes.c_int, IntArray, ctypes.POINTER(ctypes.c_int)]
    ret = lib.qstate_measure(ctypes.byref(qs),
                             ctypes.c_double(angle), ctypes.c_double(phase),
                             ctypes.c_int(qubit_num), qid_array, ctypes.byref(c_mval))

    if ret == cfg.FALSE:
        raise ValueError("can't measure the qubits.")

    mval = c_mval.value
    digits = len(qid)
    mval = '{:0{digits}b}'.format(mval, digits=digits)

    return mval

def qstate_measure_stats(qs, qid=None, shots=cfg.DEF_SHOTS, angle=0.0, phase=0.0):
    """ measurement of the qubits and get stats """

    if qid is None or qid == []:
        qid = list(range(qs.qubit_num))

    # error check
    qstate_check_args(qs, kind=cfg.MEASURE, qid=qid)

    # operate
    qubit_num = len(qid)
    qubit_id = [0 for _ in range(qubit_num)]
    for i, q in enumerate(qid):
        qubit_id[i] = q
    IntArray = ctypes.c_int * qubit_num
    qid_array = IntArray(*qubit_id)

    mdata = None
    c_mdata = ctypes.c_void_p(mdata)

    lib.qstate_measure_stats.restype = ctypes.c_int
    lib.qstate_measure_stats.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                         ctypes.c_double, ctypes.c_double,
                                         ctypes.c_int, IntArray,
                                         ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qstate_measure_stats(ctypes.byref(qs), ctypes.c_int(shots),
                                   ctypes.c_double(angle), ctypes.c_double(phase),
                                   ctypes.c_int(qubit_num), qid_array, c_mdata)

    if ret == cfg.FALSE:
        raise ValueError("can't measure the qubits.")

    out = ctypes.cast(c_mdata.value, ctypes.POINTER(MData))

    return out.contents

def qstate_measure_bell_stats(qs, qid=None, shots=cfg.DEF_SHOTS):
    """ bell measurement of the qubits and get stats """

    if qid is None or qid == []:
        qid = list(range(2))

    # error check
    qstate_check_args(qs, kind=cfg.MEASURE_BELL, qid=qid)

    # operate
    qubit_num = 2
    qubit_id = [0 for _ in range(qubit_num)]
    for i in range(qubit_num):
        qubit_id[i] = qid[i]
    IntArray = ctypes.c_int * qubit_num
    qid_array = IntArray(*qubit_id)

    mdata = None
    c_mdata = ctypes.c_void_p(mdata)

    lib.qstate_measure_bell_stats.restype = ctypes.c_int
    lib.qstate_measure_bell_stats.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                              ctypes.c_int, IntArray,
                                              ctypes.POINTER(ctypes.c_void_p)]

    ret = lib.qstate_measure_bell_stats(ctypes.byref(qs), ctypes.c_int(shots),
                                        ctypes.c_int(qubit_num), qid_array, c_mdata)

    if ret == cfg.FALSE:
        raise ValueError("can't measure the qubits.")

    out = ctypes.cast(c_mdata.value, ctypes.POINTER(MData))

    return out.contents

def qstate_operate_qcirc(qstate, cmem, qcirc, shots, cid):
    """ operate quantum circuit """

    lib.qstate_operate_qcirc.restype = ctypes.c_int
    lib.qstate_operate_qcirc.argtypes = [ctypes.POINTER(QState),
                                         ctypes.POINTER(CMem), ctypes.POINTER(QCirc)]

    if cmem is not None:

        cmem_num = cmem.cmem_num
        frequency = Counter()
        for n in range(shots):

            if n < shots - 1:
                qstate_tmp = qstate.clone()
                cmem_tmp = cmem.clone()
                ret = lib.qstate_operate_qcirc(ctypes.byref(qstate_tmp),
                                               ctypes.byref(cmem_tmp), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem_tmp.bit_array, ctypes.POINTER(ctypes.c_ubyte*cmem_num))
            else:
                ret = lib.qstate_operate_qcirc(ctypes.byref(qstate), ctypes.byref(cmem),
                                               ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem.bit_array, ctypes.POINTER(ctypes.c_ubyte*cmem_num))

            if ret == cfg.FALSE:
                raise ValueError("can't operate the quantum circuit.")

            cmem_list = [bit_array.contents[i] for i in range(cmem_num)]
            cmem_list_part = [cmem_list[c] for c in cid]
            mval = "".join(map(str, cmem_list_part))

            frequency[mval] += 1

        return frequency

    c_cmem = ctypes.POINTER(CMem)()
    ret = lib.qstate_operate_qcirc(ctypes.byref(qstate), c_cmem, ctypes.byref(qcirc))

    if ret == cfg.FALSE:
        raise ValueError("can't operate the quantum circuit.")

    return None

def qstate_free(qs):
    """ free memory of the QState object """

    lib.qstate_free.argtypes = [ctypes.POINTER(QState)]
    lib.qstate_free(ctypes.byref(qs))



# -*- coding: utf-8 -*-
""" wrapper functions for MData """
import ctypes
from ctypes.util import find_library
import pathlib

import qlazy.config as cfg
from qlazy.util import get_lib_ext
from qlazy.MData import MData

lib = ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"), mode=ctypes.RTLD_GLOBAL)

def mdata_init(qubit_num=None, shots=1, angle=0.0, phase=0.0, qid=None):
    """ initialize MData object """

    if qid is None or qid == []:
        qid = list(range(qubit_num))

    IntArray = ctypes.c_int * cfg.MAX_QUBIT_NUM
    qid_array = IntArray(*qid)

    mdata = None
    c_mdata = ctypes.c_void_p(mdata)

    lib.mdata_init.restype = ctypes.c_int
    lib.mdata_init.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_double, ctypes.c_double,
                               IntArray, ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.mdata_init(ctypes.c_int(qubit_num), ctypes.c_int(shots),
                         ctypes.c_double(angle), ctypes.c_double(phase), qid_array, c_mdata)

    if ret == cfg.FALSE:
        raise ValueError("can't initialize MData object.")

    return c_mdata


def mdata_print(md):
    """ print MData object """

    lib.mdata_print.restype = ctypes.c_int
    lib.mdata_print.argtypes = [ctypes.POINTER(MData)]
    ret = lib.mdata_print(ctypes.byref(md))

    if ret == cfg.FALSE:
        raise ValueError("can't print MData object.")

def mdata_free(md):
    """ free memory of MData object """

    lib.mdata_free.argtypes = [ctypes.POINTER(MData)]
    lib.mdata_free(ctypes.byref(md))



# -*- coding: utf-8 -*-
""" wrapper functions for Observable """
import ctypes
from ctypes.util import find_library
import pathlib

import qlazy.config as cfg
from qlazy.util import get_lib_ext
from qlazy.Observable import Observable

lib = ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"), mode=ctypes.RTLD_GLOBAL)

def observable_init(ostr):
    """ initialize Observable object """

    c_str = ctypes.create_string_buffer(ostr.encode('utf-8'))

    observ = None
    c_observ = ctypes.c_void_p(observ)

    lib.observable_init.restype = ctypes.c_int
    lib.observable_init.argtypes = [ctypes.POINTER(ctypes.c_char),
                                    ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.observable_init(c_str, c_observ)

    if ret == cfg.FALSE:
        raise ValueError("can't initialize Observable object.")

    return c_observ

def observable_free(ob):
    """ free memory of Observable object """

    lib.observable_free.argtypes = [ctypes.POINTER(Observable)]
    lib.observable_free(ctypes.byref(ob))



# -*- coding: utf-8 -*-
""" run function for qulacs's cpu/gpu simulator """

from collections import Counter
import cmath
import numpy as np

from qulacs import QuantumState
from qulacs import QuantumCircuit
from qulacs.gate import X, Y, Z
from qulacs.gate import H, S, Sdag, T, Tdag, sqrtX, sqrtXdag
from qulacs.gate import CNOT, CZ, SWAP
from qulacs.gate import RX, RY, RZ
from qulacs.gate import U1, U2, U3
from qulacs.gate import Measurement
from qulacs.gate import DenseMatrix
from qulacs.gate import to_matrix_gate

import qlazy.config as cfg
from qlazy.util import get_qgate_qubit_num, get_qgate_param_num
from qlazy.Result import Result

GateFunctionName = {
    # 1-qubit, 0-parameter gate
    cfg.PAULI_X: 'X',
    cfg.PAULI_Y: 'Y',
    cfg.PAULI_Z: 'Z',
    cfg.ROOT_PAULI_X:'sqrtX',
    cfg.ROOT_PAULI_X_:'sqrtXdag',
    cfg.HADAMARD:'H',
    cfg.PHASE_SHIFT_S:'S',
    cfg.PHASE_SHIFT_S_:'Sdag',
    cfg.PHASE_SHIFT_T:'T',
    cfg.PHASE_SHIFT_T_:'Tdag',
    cfg.IDENTITY:'Identity',
    # 1-qubit, 1-parameter gate
    cfg.ROTATION_X:'RX',
    cfg.ROTATION_Y:'RY',
    cfg.ROTATION_Z:'RZ',
    cfg.PHASE_SHIFT:'__get_P',
    cfg.ROTATION_U1:'U1',
    # 1-qubit, 2-parameter gate
    cfg.ROTATION_U2:'U2',
    # 1-qubit, 3-parameter gate
    cfg.ROTATION_U3:'U3',
    # 2-qubit, 0-parameters gate
    cfg.CONTROLLED_X:'CNOT',
    cfg.CONTROLLED_Y:'__get_CY',
    cfg.CONTROLLED_Z:'CZ',
    cfg.CONTROLLED_XR:'__get_CXR',
    cfg.CONTROLLED_XR_:'__get_CXR_dg',
    cfg.CONTROLLED_H:'__get_CH',
    cfg.CONTROLLED_S:'__get_CS',
    cfg.CONTROLLED_S_:'__get_CS_dg',
    cfg.CONTROLLED_T:'__get_CT',
    cfg.CONTROLLED_T_:'__get_CT_dg',
    cfg.CONTROLLED_P:'__get_CP',
    cfg.SWAP_QUBITS:'SWAP',
    # 2-qubit, 1-parameters gate
    cfg.CONTROLLED_RX:'__get_CRX',
    cfg.CONTROLLED_RY:'__get_CRY',
    cfg.CONTROLLED_RZ:'__get_CRZ',
    cfg.CONTROLLED_U1:'__get_CU1',
    # 2-qubit, 2-parameters gate
    cfg.CONTROLLED_U2:'__get_CU2',
    # 2-qubit, 3-parameters gate
    cfg.CONTROLLED_U3:'__get_CU3',
}

def run_cpu(qcirc=None, shots=1, cid=None, backend=None):
    """ run the quantum circuit (CPU) """

    return __run_all(qcirc=qcirc, shots=shots, cid=cid, backend=backend, proc='CPU')

def run_gpu(qcirc=None, shots=1, cid=None, backend=None):
    """ run the quantum circuit (GPU) """

    return __run_all(qcirc=qcirc, shots=shots, cid=cid, backend=backend, proc='GPU')

def __run_all(qcirc=None, shots=1, cid=None, backend=None, proc='CPU'):

    if qcirc is None:
        raise ValueError("quantum circuit must be specified.")

    qubit_num = qcirc.qubit_num
    cmem_num = qcirc.cmem_num

    if cid is None:
        cid = list(range(cmem_num))

    if cmem_num < len(cid):
        raise ValueError("length of cid must be less than classical resister size of qcirc")

    #
    # initialize
    #

    if proc == 'CPU':
        qstate = QuantumState(qubit_num)
    else:
        from qulacs import QuantumStateGpu
        qstate = QuantumStateGpu(qubit_num)
    cmem = [0] * cmem_num

    #
    # before measurement gate
    #

    while True:
        kind = qcirc.kind_first()
        if kind is None or kind is cfg.MEASURE or kind is cfg.RESET:
            break

        (kind, qid, para, c, ctrl) = qcirc.pop_gate()
        if ctrl is None or (ctrl is not None and cmem[ctrl] == 1):
            __qulacs_operate_qgate(qstate, qubit_num, kind=kind, qid=qid,
                                   phase=para[0], phase1=para[1], phase2=para[2])

    if kind is None:
        info = {'quantumstate': qstate, 'cmem': cmem}

        result = Result()
        result.qubit_num = qubit_num
        result.cmem_num = cmem_num
        result.cid = cid
        result.shots = shots
        result.frequency = None
        result.backend = backend
        result.info = info
        return result

    #
    # after measurement gate
    #

    frequency = Counter()
    qstate_tmp = None
    for _ in range(shots):

        qstate_tmp = qstate.copy()
        qcirc_tmp = qcirc.clone()

        while True:

            kind = qcirc_tmp.kind_first()
            if kind is None:
                break

            # elif kind == cfg.MEASURE:
            if kind == cfg.MEASURE:
                (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()
                mval = __qulacs_measure(qstate_tmp, qubit_num, qid[0])
                if c is not None:
                    cmem[c] = mval

            elif kind == cfg.RESET:
                (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()
                __qulacs_reset(qstate_tmp, qubit_num, qid[0])

            else:
                (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()
                if (ctrl is None or (ctrl is not None and cmem[ctrl] == 1)):
                    __qulacs_operate_qgate(qstate_tmp, qubit_num, kind=kind, qid=qid,
                                           phase=para[0], phase1=para[1], phase2=para[2])

        if len(cmem) > 0:
            mval = ''.join(map(str, [cmem[i] for i in cid]))
            frequency[mval] += 1

    if qstate_tmp is not None:
        qstate.load(qstate_tmp.get_vector())

    if len(frequency) == 0:
        frequency = None

    info = {'quantumstate': qstate, 'cmem': cmem}

    result = Result()
    result.backend = backend
    result.qubit_num = qubit_num
    result.cmem_num = cmem_num
    result.cid = cid
    result.shots = shots
    result.frequency = frequency
    result.info = info

    return result

def __is_supported_qgate(kind):

    return kind in GateFunctionName.keys()

# not supported as pre-defined gates

def __get_P(q0, phase):

    exp = cmath.exp(1.j * phase)
    gate = DenseMatrix(q0, [[1., 0.], [0., exp]])
    return gate

def __get_CY(q0, q1):

    gate = to_matrix_gate(Y(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CXR(q0, q1):

    gate = to_matrix_gate(sqrtX(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CXR_dg(q0, q1):

    gate = to_matrix_gate(sqrtXdag(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CH(q0, q1):

    gate = to_matrix_gate(H(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CS(q0, q1):

    gate = to_matrix_gate(S(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CS_dg(q0, q1):

    gate = to_matrix_gate(Sdag(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CT(q0, q1):

    gate = to_matrix_gate(T(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CT_dg(q0, q1):

    gate = to_matrix_gate(Tdag(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CP(q0, q1, phase):

    exp = cmath.exp(1.j * phase)
    gate = DenseMatrix(q1, [[1., 0.], [0., exp]])
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CRX(q0, q1, phase):

    gate = to_matrix_gate(RX(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CRY(q0, q1, phase):

    gate = to_matrix_gate(RY(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CRZ(q0, q1, phase):

    gate = to_matrix_gate(RZ(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CU1(q0, q1, phase):

    gate = to_matrix_gate(U1(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CU2(q0, q1, phase, phase1):

    gate = to_matrix_gate(U2(q1, phase, phase1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CU3(q0, q1, phase, phase1, phase2):

    gate = to_matrix_gate(U3(q1, phase, phase1, phase2))
    gate.add_control_qubit(q0, 1)
    return gate

def __qulacs_operate_qgate(qstate, qubit_num, kind, qid, phase, phase1, phase2):

    if __is_supported_qgate(kind) is False:
        raise ValueError("not supported quantum gate")

    circ = QuantumCircuit(qubit_num)

    term_num = get_qgate_qubit_num(kind)
    para_num = get_qgate_param_num(kind)
    gate_function_name = GateFunctionName[kind]

    phase = phase * np.pi
    phase1 = phase1 * np.pi
    phase2 = phase2 * np.pi

    # the sign-definition of rotation gate on qulacs
    if (kind in (cfg.ROTATION_X, cfg.ROTATION_Y, cfg.ROTATION_Z,
                 cfg.CONTROLLED_RX, cfg.CONTROLLED_RY, cfg.CONTROLLED_RZ)):
        phase = -phase
    # the argument-order-definition of U2 gate on qulacs
    elif kind in (cfg.ROTATION_U2, cfg.CONTROLLED_U2):
        phase1, phase = phase, phase1
    # the argument-order-definition of U3 gate on qulacs
    elif kind in (cfg.ROTATION_U3, cfg.CONTROLLED_U3):
        phase2, phase = phase, phase2

    if term_num == 1 and para_num == 0:
        circ.add_gate(eval(gate_function_name)(qid[0]))
    elif term_num == 1 and para_num == 1:
        circ.add_gate(eval(gate_function_name)(qid[0], phase))
    elif term_num == 1 and para_num == 2:
        circ.add_gate(eval(gate_function_name)(qid[0], phase, phase1))
    elif term_num == 1 and para_num == 3:
        circ.add_gate(eval(gate_function_name)(qid[0], phase, phase1, phase2))
    elif term_num == 2 and para_num == 0:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1]))
    elif term_num == 2 and para_num == 1:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1], phase))
    elif term_num == 2 and para_num == 2:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1], phase, phase1))
    elif term_num == 2 and para_num == 3:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1], phase, phase1, phase2))
    else:
        raise ValueError("not supported terminal or parameter-number")

    circ.update_quantum_state(qstate)

def __qulacs_reset(qstate, qubit_num, q):

    # error check
    if q >= qubit_num:
        raise ValueError("reset qubit id is out of bound")

    circ = QuantumCircuit(qubit_num)
    circ.add_gate(Measurement(q, 0))
    circ.update_quantum_state(qstate)
    circ_flip = QuantumCircuit(qubit_num)
    if qstate.get_classical_value(0) == 1:
        circ_flip.add_gate(X(q))
        circ_flip.update_quantum_state(qstate)

def __qulacs_measure(qstate, qubit_num, q):

    # error check
    if q >= qubit_num:
        raise ValueError("measurement qubit id is out of bound")

    circ = QuantumCircuit(qubit_num)
    circ.add_gate(Measurement(q, 0))
    circ.update_quantum_state(qstate)
    mval = qstate.get_classical_value(0)

    return mval
